from pathlib import Path
import datetime as dt
from typing import Optional, Union

import requests
import numpy as np
import pandas as pd
import xarray as xr
from netCDF4 import Dataset  # type: ignore
from scipy.interpolate import interp1d

from gfatpy.utils.utils import parse_datetime

API_URL = "https://cloudnet.fmi.fi/api"
BASE_QUERY_PARAMS = {"site": "granada"}


def get_ecmwf_day(
    date: dt.datetime | str, times: np.ndarray, heights: np.ndarray
) -> xr.Dataset:
    """Fetches the ECMWF API to retrieve temperature and pressure data

    Args:
        date (dt.datetime | str): A data, hour doesn't affect

    Raises:
        FileNotFoundError: There is no available data in the ECMWF data

    Returns:
        xr.Dataset: x array dataset with variables: ["pressure", "temperature", "height", "time"]
    """
    _date = parse_datetime(date)

    response_filename = requests.get(
        f"{API_URL}/model-files",
        params={**BASE_QUERY_PARAMS, "date": _date.date().isoformat()},
    )

    if not response_filename.ok:
        raise RuntimeError(
            f"Request returned code {response_filename.status_code}: {response_filename.text}"
        )

    response_filename = response_filename.json()

    if len(response_filename) == 0:
        raise FileNotFoundError("No data found for the given day")

    file = requests.get(response_filename[0]["downloadUrl"])
    nc = Dataset("ecmwf.nc", memory=file.content)
    dataset = xr.open_dataset(xr.backends.NetCDF4DataStore(nc))  # type: ignore

    if heights is None:
        heights = np.arange(2666) * 7.5

    d_height = np.tile(heights, (25, 1))
    pressure = np.empty_like(d_height)
    temperature = np.empty_like(d_height)

    hours = np.arange(25)

    for hour in hours:
        pressure[hour, :] = np.interp(
            heights, nc["height"][hour, :], nc["pressure"][hour, :]
        )
        temperature[hour, :] = np.interp(
            heights, nc["height"][hour, :], nc["temperature"][hour, :]
        )

    pressure = np.apply_along_axis(
        lambda row: np.interp(
            times.astype("float64"), dataset.time.values.astype("float64"), row
        ).T,
        0,
        pressure,
    )

    temperature = np.apply_along_axis(
        lambda row: np.interp(
            times.astype("float64"), dataset.time.values.astype("float64"), row
        ).T,
        0,
        temperature,
    )

    return xr.Dataset(
        {
            "temperature": (["time", "range"], temperature),
            "pressure": (["time", "range"], pressure),
        },
        coords={"range": heights, "time": times},
    )  # type: ignore


def get_ecmwf_temperature_preasure(
    date: dt.datetime | str,
    hour: int | None = None,
    heights: np.ndarray | list | None = None,
) -> pd.DataFrame:
    """Fetch the ecmwf API and downloads temperature and pressure for a given hour date and hour (optional and default to 12:00). Returns the information in a DataFrame

    Args:
        date (dt.datetime | str): datetime or str datetime
        hour (int | None, optional): If need an specific time for the data retieval. Defaults to None.
        heights (np.ndarray | list | None, optional): A heights profile . Defaults to None.

    Raises:
        FileNotFoundError: If ecmwf model doesn't have data available for that date
        ValueError: Raises in case ranges input array shape is not as expected

    Returns:
        pd.DataFrame: DataFrame with three columns: height, temperature, pressure
    """

    _date = parse_datetime(date)

    if hour is None:
        _date = _date.replace(hour=12)
    else:
        _date = _date.replace(hour=hour)

    hour_int: int = _date.hour

    response_filename = requests.get(
        f"{API_URL}/model-files",
        params={**BASE_QUERY_PARAMS, "date": _date.date().isoformat()},
    ).json()

    if len(response_filename) == 0:
        raise FileNotFoundError("No data found for the given day")

    file = requests.get(response_filename[0]["downloadUrl"])
    nc = Dataset("ecmwf.nc", memory=file.content)

    height = nc.variables["height"][hour_int]
    temperature = nc.variables["temperature"][hour_int]
    pressure = nc.variables["pressure"][hour_int]

    if heights is None:
        return pd.DataFrame(
            {"height": height, "temperature": temperature, "pressure": pressure}
        )

    if np.ndim(heights) != 1:
        raise ValueError("heights must be 1-dimensinal")

    # interpolations, case when range doesn't match API heights
    temperature_fn = interp1d(
        height, temperature, bounds_error=False, fill_value="extrapolate"  # type: ignore
    )
    pressure_fn = interp1d(
        height, pressure, bounds_error=False, fill_value="extrapolate"  # type: ignore
    )

    return pd.DataFrame(
        {
            "height": heights,
            "temperature": temperature_fn(heights),
            "pressure": pressure_fn(heights),
        }
    )




def get_data_as_ascii(
    target_datetime: dt,    
    output_dir: Path | str | None = None
) -> Path:
    """
    Fetches temperature and pressure data for a given target datetime,
    saves the data in ASCII format to a file, and returns the path of the saved file.

    Args:
        target_datetime (datetime.datetime): The target date and time.
        output_dir (Path | str | None, optional): The directory where the output file
            will be saved. If not provided, the current working directory is used.

    Raises:
        ValueError: If the output_dir is not a valid directory.
        FileNotFoundError: If no data is found for the given day.

    Returns:
        Path: The path of the saved ASCII file.
    """
    if output_dir is None:
        output_dir = Path.cwd()
    elif isinstance(output_dir, str):
        output_dir = Path(output_dir)

    if not output_dir.exists() or not output_dir.is_dir():
        raise ValueError(f"output_dir must be a valid directory: {output_dir}")

    hour_int: int = target_datetime.hour

    response_filename = requests.get(
        f"{API_URL}/model-files",
        params={**BASE_QUERY_PARAMS, "date": target_datetime.date().isoformat()},
    ).json()

    if len(response_filename) == 0:
        raise FileNotFoundError("No data found for the given day")

    # Fetch temperature and pressure data
    file = requests.get(response_filename[0]["downloadUrl"])
    nc = Dataset(output_dir / "ecmwf.nc", memory=file.content)

    height = nc.variables["height"][hour_int]
    temperature = nc.variables["temperature"][hour_int]
    pressure = nc.variables["pressure"][hour_int]

    # Generate ASCII content
    ascii_content = ""
    for idx, h in enumerate(height):
        t = temperature[idx]
        p = pressure[idx]/100.0
        line = f"{h:.2f}\t{t:.2f}\t{p:.2f}\n"
        ascii_content += line

    # Create filename using target_datetime
    current_datetime = target_datetime.strftime("%Y%m%d_%H%M")
    output_file = output_dir / f"{current_datetime}_ecmwf.txt"

    # Save ASCII content to file
    with open(output_file, "w") as file_:
        file_.write("height[m] temperature[K] pressure[hPa]\n")
        file_.write(ascii_content)

    return output_file


all = ["get_ecmwf_day", "get_ecmwf_temperature_preasure", "get_data_as_ascii"]



from pdb import set_trace
import numpy as np
from datetime import datetime

from scipy.integrate import cumulative_trapezoid, cumtrapz
from scipy.constants import Boltzmann

from gfatpy.atmo import atmo, ecmwf
from gfatpy.lidar.types import ParamsDict
from gfatpy.lidar.utils import extrapolate_beta_with_angstrom
from gfatpy.lidar.utils import sigmoid


def generate_particle_properties(
    ranges: np.ndarray,
    wavelength: float,
    ae: float | tuple[float, float] = (1.5, 0),
    lr: float | tuple[float, float] = (75, 45),
    synthetic_beta: float | tuple[float, float] = (2.5e-6, 2.0e-6),
) -> np.ndarray:
    """_summary_

    Args:
        ranges (np.ndarray): ranges
        wavelength (float): wavelength
        fine_ae (float): fine-mode Angstrom exponent
        coarse_ae (float): coarse-mode Angstrom exponent
        fine_beta532 (float, optional): fine-mode backscatter coefficient at 532 nm. Defaults to 2.5e-6.
        coarse_beta532 (float, optional): coarse-mode backscatter coefficient at 532 nm. Defaults to 2.0e-6.

    Returns:
        np.ndarray: particle backscatter coefficient profile
    """
    if isinstance(ae, tuple):
        fine_ae = ae[0]
        coarse_ae = ae[1]
    else:
        fine_ae = ae
        coarse_ae = ae

    if isinstance(lr, tuple):
        fine_lr = lr[0]
        coarse_lr = lr[1]
    else:
        fine_lr = lr
        coarse_lr = lr

    if isinstance(synthetic_beta, tuple):
        fine_beta532 = synthetic_beta[0]
        coarse_beta532 = synthetic_beta[1]
    else:
        fine_beta532 = synthetic_beta
        coarse_beta532 = synthetic_beta

    beta_part_fine_532 = sigmoid(
        ranges, 2500, 1 / 60, coeff=-fine_beta532, offset=fine_beta532
    )
    beta_part_coarse_532 = sigmoid(
        ranges, 5000, 1 / 60, coeff=-coarse_beta532, offset=coarse_beta532
    )

    beta_part_fine = extrapolate_beta_with_angstrom(
        beta_part_fine_532, 532, wavelength, fine_ae
    )

    beta_part_coarse = extrapolate_beta_with_angstrom(beta_part_coarse_532, 532, wavelength, coarse_ae)

    beta_total = beta_part_fine + beta_part_coarse

    alpha_part_fine = fine_lr * beta_part_fine
    alpha_part_coarse = coarse_lr * beta_part_coarse

    alpha_total = alpha_part_fine + alpha_part_coarse

    return beta_part_fine, beta_part_coarse, beta_total, alpha_part_fine, alpha_part_coarse, alpha_total


def synthetic_signals(
    ranges: np.ndarray,
    wavelengths: float | tuple[float, float] = 532,
    wavelength_raman: float | None = None,
    overlap_midpoint: float = 600,
    k_lidar: float | tuple[float, float] = (1e10, 1e9),
    ae: float | tuple[float, float] = (1.5, 0),
    lr: float | tuple[float, float] = (75, 45),
    synthetic_beta: float | tuple[float, float] = (2.5e-6, 2.0e-6),
    force_zero_aer_after_bin: int | None = None,
    paralell_perpendicular_ratio: float = 0.33,
    meteo_profiles: tuple[np.ndarray, np.ndarray] | None = None,    
    apply_overlap: bool = True,    
) -> tuple[np.ndarray, np.ndarray | None, ParamsDict]:
    """It generates synthetic lidar signal.

    Args:
        ranges (np.ndarray): Range
        wavelength (float, optional): Wavelength. Defaults to 532.
        overlap_midpoint (float, optional): _description_. Defaults to 600.
        k_lidar (float, optional): Lidar constant calibration. Defaults to 4e9.
        wavelength_raman (float | None, optional): Raman wavelength. Defaults to None. If None, signal is elastic.
        paralell_perpendicular_ratio (float, optional): _description_. Defaults to 0.33.
        particle_lratio (float, optional): _description_. Defaults to 45.
        force_zero_aer_after_bin (int | None, optional): _description_. Defaults to None.

    Returns:
        tuple[np.ndarray, ParamsDict]: _description_
    """

    z = ranges

    # Overlap
    if apply_overlap:
        overlap = sigmoid(
            z,
            overlap_midpoint,
            1 / 50,
            offset=0,
        )

        overlap[overlap < 9e-3] = 0
        overlap[overlap > 0.999] = 1
    else:
        overlap = 1 

    if isinstance(lr, float):
        lr = (lr, lr)
    if isinstance(ae, float):
        ae = (ae, ae)
    if isinstance(synthetic_beta, float):
        synthetic_beta = (synthetic_beta, synthetic_beta)
    if isinstance(k_lidar, float):
        k_lidar_elastic = k_lidar
    else:
        k_lidar_elastic, k_lidar_raman = k_lidar

    #Check temperature and pressure profiles
    if meteo_profiles is None:
        ecmwf_data = ecmwf.get_ecmwf_temperature_preasure(datetime(2022, 9, 3), heights=z)
        P = ecmwf_data.pressure.values
        T = ecmwf_data.temperature.values
    else:
        #check length of meteo_profiles with z
        if len(meteo_profiles[0]) != len(z):
            raise ValueError('Length of meteo_profiles must be equal to length of z')
        else:
            P = meteo_profiles[0]
            T = meteo_profiles[1]

    #Check if wavelength is a tuple
    if isinstance(wavelengths, tuple):
        wavelength = wavelengths[0]
        wavelength_raman = wavelengths[1]
    else:
        wavelength = wavelengths        
        wavelength_raman = None
    
    #Generate molecular profiles for elastic wavelength
    atmo_data = atmo.molecular_properties(
        wavelength,
        P,
        T,
        heights=z,
    )
    beta_mol = atmo_data["molecular_beta"].values
    alpha_mol = atmo_data["molecular_alpha"].values
            
    #TODO
    # Particle elastic 
    # 0.33 para polvo desértico
    # 0.0034 para parte molecular
    # Calcular beta_part parallel = total*(1-0.33)
    # Calcular beta_part perpendicular = total*(0.33)
    # Análogo con otro coeficiente para la parte molecular

    _, _, beta_part, alpha_part_fine, alpha_part_coarse, alpha_part = generate_particle_properties(
        ranges, wavelength, ae=ae, lr=lr, synthetic_beta=synthetic_beta 
    ) 
        
    # Elastic transmittance
    T_elastic = np.exp(-cumulative_trapezoid(alpha_mol+ alpha_part, z, initial=0))  # type: ignore

    #Elastic signal
    P_elastic = k_lidar_elastic * (overlap / z**2) * (beta_part + beta_mol) * T_elastic**2

    #Save parameters to create synthetic elastic signal 
    params: ParamsDict = {
        "particle_beta": beta_part,
        "particle_alpha": alpha_part,
        "molecular_beta": beta_mol,
        "molecular_alpha": alpha_mol,
        "lidar_ratio" : lr,
        "molecular_beta_att": atmo_data["attenuated_molecular_beta"].values,
        "overlap": overlap,
        "k_lidar": k_lidar,
        "particle_angstrom_exponent": ae,
        "synthetic_beta": synthetic_beta,
        "temperature": T, 
        "pressure": P}

    # Raman signal
    if wavelength_raman is not None:
        # Generate molecular profiles for raman wavelength
        wavelength_raman = 607
        atmo_data_raman = atmo.molecular_properties(wavelength_raman, P, T, heights=z)
        beta_mol_raman = atmo_data_raman["molecular_beta"].values
        alpha_mol_raman = atmo_data_raman["molecular_alpha"].values

        # Alpha particle raman
        alpha_part_fine_raman = alpha_part_fine *(wavelength_raman/wavelength)**(-ae[0])
        alpha_part_coarse_raman = alpha_part_coarse *(wavelength_raman/wavelength)**(-ae[1])
        alpha_part_raman = alpha_part_fine_raman + alpha_part_coarse_raman

        # Molecule density    
        N = atmo.number_density_at_pt(P, T)

        # Transmittance Raman
        T_raman = np.exp(-cumulative_trapezoid(alpha_mol_raman + alpha_part_raman, z, initial=0))  # type: ignore

        P_raman = k_lidar_raman * (overlap / z**2) * beta_mol_raman * T_elastic * T_raman

        params["molecular_alpha_raman"] = alpha_mol_raman
        params["molecular_beta_raman"] = beta_mol_raman
        params["molecular_beta_att_raman"] = atmo_data_raman["attenuated_molecular_beta"].values,        
        params["transmittance_raman"] = T_raman
        params["overlap"] = overlap
        params['molecular_density'] = N

    else:
        P_raman = None

    if force_zero_aer_after_bin is not None:
        alpha_part[force_zero_aer_after_bin:] = 0
        beta_part[force_zero_aer_after_bin:] = 0

    return P_elastic, P_raman, params



import pathlib
import numpy as np
import pandas as pd
from datetime import datetime
from scipy import integrate
from scipy.interpolate import interp1d
from loguru import logger

from gfatpy.utils.io import read_yaml

# AERONET SYSTEM INFO
AERONET_INFO_FILE = pathlib.Path(__file__).parent.absolute() / "info.yml"
AERONET_INFO = read_yaml(AERONET_INFO_FILE)


def find_available_aod_wavelengths(header: dict, df: pd.DataFrame) -> list:
    """find available aod wavelengths in AERONET dataframe.

    Args:
        header (dict): AERONET header
        df (pd.DataFrame): AERONET dataframe
    Returns:
        list[int]: available wavelenghts
    """

    available_wavelengths = []
    aeronet_data = header["aeronet_data"].replace(" ", "_")
    for wave_ in AERONET_INFO["wavelengths"][aeronet_data]:
        property2check = AERONET_INFO["aod_string_type"][aeronet_data].replace(
            "&", f"{wave_}"
        )
        if property2check in df.keys():
            available_wavelengths.append(wave_)
    available_wavelengths = np.sort(np.asarray(available_wavelengths)).tolist()
    return available_wavelengths


def find_nearest_wavelength(
    header: dict, df: pd.DataFrame, target_wavelength: float
) -> list[float]:
    """Find the available wavelength above [`larger_wavelength`] and below [`lower_wavelength`] of the `target_wavelength`.

    Args:
        header (dict): AERONET header.
        df (pd.DataFrame): AERONET dataframe.
        target_wavelength (float): wavelength to be interpolated.

    Returns:
        list[float]: lower_wavelength, larger_wavelength
    """

    available_wavelengths = find_available_aod_wavelengths(header, df)
    idx_to_be_placed = np.searchsorted(
        available_wavelengths, [target_wavelength], side="left"
    )[0]
    if idx_to_be_placed == 0:  # Extrapolation is required
        bottom_wave = available_wavelengths[0]
        up_wave = available_wavelengths[1]
    elif idx_to_be_placed == len(available_wavelengths):  # Extrapolation is required
        bottom_wave = available_wavelengths[-2]
        up_wave = available_wavelengths[-1]
    else:
        bottom_wave = available_wavelengths[idx_to_be_placed - 1]
        up_wave = available_wavelengths[idx_to_be_placed + 1]
    return bottom_wave, up_wave


def add_interpol_aod(
    header: dict,
    df: pd.DataFrame,
    target_wavelength: int,
    bottom_wavelength: int | None = None,
    up_wavelength: int | None = None,
    allow_search: bool = True,
) -> pd.DataFrame:
    """Add the interpolated aod of `target_wavelength` to the AERONET dataframe.

    Args:
        header (dict): AERONET header.
        df (pd.DataFrame): AERONET dataframe.
        target_wavelength (int): Target wavelength
        bottom_wavelength (int | None, optional): lower wavelength to make the interpolation. Defaults to None. With None, the nearest lower wavelength available is used.
        up_wavelength (int | None, optional): lower wavelength to make the interpolation. Defaults to None. With None, the nearest larger wavelength available is used.
        allow_search (bool | optional): allow to search the best lower and larger wavelength if the provided are not available.
    Raises:
        ValueError: Interpolation not possible since bottom/up wavelengths are not available.

    Returns:
        pd.DataFrame: AERONET dataframe with new interpolated AOD at `target_wavelength`.
    """
    available_wavelengths = find_available_aod_wavelengths(header, df)

    if (
        bottom_wavelength not in available_wavelengths
        or up_wavelength not in available_wavelengths
    ) and allow_search is False:
        logger.error(
            "Interpolation not possible since bottom/up wavelengths are not available."
        )
        raise ValueError

    if bottom_wavelength is None or up_wavelength is None:
        bottom_wavelength, up_wavelength = find_nearest_wavelength(
            header, df, target_wavelength
        )
    # Property strings
    bottom_property = AERONET_INFO["aod_string_type"][
        header["aeronet_data"].replace(" ", "_")
    ].replace("&", f"{bottom_wavelength}")

    up_property = AERONET_INFO["aod_string_type"][
        header["aeronet_data"].replace(" ", "_")
    ].replace("&", f"{up_wavelength}")

    target_property = bottom_property.replace(
        f"{bottom_wavelength}", f"{target_wavelength}"
    )

    # Retrieve angstrom exponent
    angstrom_exponent = -np.log(
        df[up_property].values / df[bottom_property].values
    ) / np.log(up_wavelength / bottom_wavelength)

    # Interpolation
    if target_wavelength > bottom_wavelength and target_wavelength < up_wavelength:
        df[target_property] = df[bottom_property].values * (
            target_wavelength / bottom_wavelength
        ) ** (-angstrom_exponent)

    # Left-side extrapolation
    if target_wavelength < bottom_wavelength:
        logger.warning("Extrapolation is used.")
        df[target_property] = df[bottom_property].values * (
            target_wavelength / bottom_wavelength
        ) ** (-angstrom_exponent)

    # Right-side extrapolation
    if target_wavelength > up_wavelength:
        logger.warning("Extrapolation is used.")
        df[target_property] = df[up_property].values * (
            target_wavelength / up_wavelength
        ) ** (-angstrom_exponent)

    return df


def add_logAOD(header: dict, df: pd.DataFrame) -> pd.DataFrame:
    """Add colums in the DataFrame with natural logarithyms of AODs.

    Args:
        df (pd.DataFrame): Aeronet DataFrame from gfatpy.aeronet.reader()

    Returns:
        df (pd.DataFrame):  Aeronet DataFrame with new columns
    """

    # Ln applied to AOD
    waves = find_available_aod_wavelengths(header, df)
    for wave_ in waves:
        aod_var = AERONET_INFO["aod_string_type"][
            header["aeronet_data"].replace(" ", "_")
        ].replace("&", f"{wave_}")

        df["lnAOD%d" % wave_] = np.log(df[aod_var])
    return df


def resample_logradius_distribution(
    distribution: np.ndarray | pd.Series,
    factor: int = 10,
    interpolation_kind: str = "quadratic",
) -> tuple[np.ndarray]:
    """Resample a logaritmic distribution.

    Args:
        distribution (np.ndarray): Aeronet DataFrame from `gfatpy.aeronet.reader()`
        factor (int, optional): Increase the resolution by dividing the current resolution by this factor. Defaults to 10.
        interpolation_kind (str, optional): Method of interpolation. Defaults to "quadratic".

    Returns:
        list[np.ndarray]:
            - Resampled distribution.
            - Resampled radius.
            - Resampled logarithmic radius.
            - New logarithmic radius resolution.
    """
    radius = np.asarray(AERONET_INFO["radius"])
    lnr = np.log(radius)
    resol_lnr = np.diff(lnr).mean()
    resample_resol_lnr = resol_lnr / factor
    resample_lnr = np.arange(lnr[0], lnr[-1] + resol_lnr / factor, resol_lnr / factor)
    resample_radius = np.exp(resample_lnr)

    # Increase resolution to improve the fitting
    resample_function = interp1d(
        radius, distribution, kind=interpolation_kind, bounds_error=True
    )
    resample_distribution_scipy = resample_function(resample_radius)
    resample_distribution_numpy = np.interp(
        resample_radius, radius, distribution.values.astype(float)
    )
    resample_distribution = np.concatenate(
        (
            resample_distribution_scipy[resample_radius < 11.432287],
            resample_distribution_numpy[resample_radius >= 11.432287],
        )
    )
    return resample_distribution, resample_radius, resample_lnr, resample_resol_lnr


def filter_df(df: pd.DataFrame, conditions_dict: dict) -> pd.DataFrame:
    """Filter the aeronet dataframe according to the conditions specified in `conditions_dict`

    Args:
        df (pd.DataFrame): Aeronet DataFrame from `gfatpy.aeronet.reader()`
        conditions_dict (dict): condition dictionary.
            Keys: Values
                'aerosol_type': ['DUST', 'HA', 'MA', 'SA', 'MIXTURE'] from `gfatpy.aeronet.classification.aerosol_type_classification_Lee()`
                'mode_predominance': [HYGHLY_COARSE, COARSE, BIMODAL, FINE, HIGHLY_FINE, UNKNOWN] from `gfatpy.aeronet.classification.mode_predominance_classification()`
                'If_Retrieval_is_L2(without_L2_0.4_AOD_440_threshold)': [1, 0]

    Returns:
        pd.DataFrame: Filtered Aeronet DataFrame
    """

    conditions = True * np.ones(len(df.index))
    for key_ in conditions_dict.keys():
        condition_ = df[key_] == conditions_dict[key_]
        conditions = np.logical_and(conditions, condition_)
    df_ = df[conditions]
    return df_


def calculate_fine_mode_fraction(
    aod_fine: np.ndarray | pd.Series,
    aod_total: np.ndarray | pd.Series,
) -> np.ndarray | pd.Series:

    """calculate_fine_mode_fraction retrieves the fine mode fraction as the fine-to-total AOD ratio.

    Args:
        aod_fine (np.ndarray): Fine AOD
        aod_total (np.ndarray): Total AOD

    Raise:
        ValueError: 'Length of arrays must be the same.'

    Returns:
        np.ndarray: fine mode fraction
    """

    if len(aod_fine) != len(aod_total):
        logger.error("Length of arrays must be the same.")
        raise ValueError

    return aod_fine / aod_total


def distribution_to_total_concentration(
    df: pd.DataFrame,
    concentration_type: str = "number",
    reference_radius_nm: int | None = None,
    integral_towards: str = "positive",
) -> pd.DataFrame:
    """Retrieve the total concentration (and natural logarithm) in two radius ranges: [r_min, r_reference] and [r_reference, r_max].

    Args:
        df (pd.DataFrame): AERONET dataframe from *.all file.
        concentration_type (str): ['surface', 'number']
        reference_radius_nm (int | None): minimum radius value to retrieve the concentration. If None, inflection radius is used.
        integral_towards (str): Direction of the integration ['positive': towards the maximum radius, 'negative': towards the minimum radius , 'both': both are retrieved and saved]. Defaults 'positive'.

    Raises:
        ValueError: Dataframe is empty.
        ValueError: integral_towardds is not one of the following options: ['positive', 'negative', 'both'].

    Returns:
        df (pd.DataFrame): dataframe with number/surface concentration and its natural logarithm 'lnN'.
    """

    if len(df) == 0:
        logger.error("Dataframe is empty.")
        raise ValueError

    # Initializing variables
    # Radius
    radius = np.asarray(AERONET_INFO["radius"])  # um¡

    # Positive
    if integral_towards == "both":
        do_positive, do_negative = True, True
    elif integral_towards == "positive":
        do_positive, do_negative = True, False
    elif integral_towards == "negative":
        do_positive, do_negative = False, True
    else:
        logger.error("Value of integral_towards not allowed.")
        raise AttributeError

        # Variables to storage concentrations
    conc_up = np.zeros(len(df))
    conc_down = np.zeros(len(df))
    if reference_radius_nm is not None:
        # Convert from nanomter to micrometer
        reference_radius_used_microm = reference_radius_nm / 1000.0
        conc_up_str = (
            f"{concentration_type}_concentration_above_{reference_radius_nm:d}"
        )
        conc_down_str = (
            f"{concentration_type}_concentration_below_{reference_radius_nm:d}"
        )
    else:
        conc_up_str = "COARSE"
        conc_down_str = "FINE"

        # Indexes of the radius in the Dataframe.
    idx_min = df.columns.get_indexer(["0.050000"])
    idx_max = df.columns.get_indexer(["15.000000"])

    # Stablising the concentration_factor (number Vs surface concentration)
    if concentration_type == "number":
        concentration_factor = 1.0
    else:
        concentration_factor = (
            4 * np.pi * radius[radius >= reference_radius_used_microm] ** 2
        )

    # Selection of the VSD
    dV_dlnr = df.iloc[:, np.arange(idx_min, idx_max + 1, 1)]  # VSD um^3/um^2
    dV_dlnr = dV_dlnr.to_numpy()

    # Retrieve NSD
    Vradius = (4.0 / 3.0) * np.pi * np.power(radius, 3)  # i-bin volume  um^3
    delta_lnr = np.diff(np.log(radius)).mean()  # resolution um
    dN_dlnr_serie = dV_dlnr / Vradius  # NSD #/um^2

    if do_positive:
        for idx in np.arange(len(df)):
            if reference_radius_nm is None:
                inflection_radius = df[
                    "Inflection_Radius_of_Size_Distribution(um)"
                ].iloc[idx]
                reference_radius_used_microm = inflection_radius / 1000.0

            dN_dlnr_ = dN_dlnr_serie[idx, :]

            conc_up[idx] = integrate.simps(
                dN_dlnr_[radius >= reference_radius_used_microm] * concentration_factor,
                dx=delta_lnr,
            )

        # Save arrays in DATAFRAME
        df[conc_up_str] = conc_up
        df[f"ln_{conc_up_str}"] = np.log(conc_up)

    if do_negative:
        for idx in np.arange(len(df)):
            if reference_radius_nm is None:
                inflection_radius = df[
                    "Inflection_Radius_of_Size_Distribution(um)"
                ].iloc[idx]
                reference_radius_used_microm = inflection_radius / 1000.0

            dN_dlnr_ = dN_dlnr_serie[idx, :]

            conc_down[idx] = integrate.simps(
                dN_dlnr_[radius < reference_radius_used_microm] * concentration_factor,
                dx=delta_lnr,
            )

        # Save arrays in DATAFRAME
        df[conc_down_str] = conc_down
        df[f"ln_{conc_down_str}"] = np.log(conc_down)
    return df


def find_aod(
    header: dict,
    df: pd.DataFrame,
    target_wavelength: int,
    target_datetime: datetime,
    allowed_time_gap_hour: int = 1,
    allowed_interpolation: bool = True,
) -> tuple[float | None, float | None]:
    """Find the aod value at `wavelength` close to the `datetime` as much as determined by `allowed_time_gap_hour`.

    Args:
        header (dict): AERONET header from `gfatpy.aeronet.reader()`
        df (pd.DataFrame): AERONET dataframe from `gfatpy.aeronet.reader()`
        target_wavelength (float): wavelength of the aod requiered. If not found, it will be interpolated with the available ones.
        target_datetime (datetime): datetime
        allowed_time_gap_hour (int, optional): allowed time gap between the mesured and required datetime. Defaults to 1. Provides None if requirement not fulfilled.
        allowed_interpolation (bool, optional):it allows wavelength interpolation.

    Raises:
        RuntimeWarning: Time gap found to be larger than allowed by the user with `allowed_time_gap_hour`
        RuntimeWarning: "Extra- inter-polation required but not allowed by the user (`allowed_interpolation` is `False`).

    Returns:
        tuple[float | None, float | None]: measured_aod, measured_datetime
    """
    """

    Args:
    Returns:
        tuple[float, datetime]:
    """
    # Set default avaiable data flag
    go = False

    # Set result to None by default
    measured_datetime, measured_aod = None, None

    # Finde the nereast datetime
    nearest_datetime = (
        df.index[df.index.get_indexer([target_datetime], method="nearest")]
    ).to_pydatetime()[0]

    # Check distance between both
    if target_datetime > nearest_datetime:
        time_gap = target_datetime - nearest_datetime
    else:
        time_gap = nearest_datetime - target_datetime
    time_gap_hour = time_gap.seconds / 60.0 / 60.0

    if time_gap_hour > allowed_time_gap_hour:
        logger.warning(
            "Time gap found to be larger than allowed by the user with `allowed_time_gap_hour`."
        )
        raise RuntimeWarning
    property = AERONET_INFO["aod_string_type"][
        header["aeronet_data"].replace(" ", "_")
    ].replace("&", f"{target_wavelength}")

    if (
        property not in df.keys() and not allowed_interpolation
    ):  # extrapolation required
        logger.warning(
            "Extra- inter-polation required but not allowed by the user (`allowed_interpolation` is `False`)."
        )
        raise RuntimeWarning

    # Extrapolation allowed
    try:
        bottom_wave, up_wave = find_nearest_wavelength(header, df, target_wavelength)
        df = add_interpol_aod(header, df, target_wavelength, bottom_wave, up_wave)
        go = True
    except Exception as e:
        logger.error("Extra- inter-polatization not finished.")
        raise RuntimeError

    if go:
        row = df.iloc[df.index.get_indexer([target_datetime], method="nearest")]
        measured_aod = row[property].values[0]
        measured_datetime = nearest_datetime

    return measured_aod, measured_datetime



from loguru import logger

DARK_SUB = {0: "not-dark-subtracted", 1: "dark-subtracted"}


def select_dc(lidar_ds, channel):
    """Read Associated Dark Current Measurements, if wanted"""
    ilu_dict = {True: "day", False: "ngt"}
    if channel[-1] == "a":
        if lidar_ds:
            dark_subtracted = lidar_ds[f"signal_{channel}"].dark_corrected
            if dark_subtracted:
                # day/ngt DC measurements
                ilu_flag = (
                    lidar_ds.cosine_sza.mean("time").values > 0.0
                )  # True: day, False: night
                dc = lidar_ds[f"dc_{channel}"].sel(ilu=ilu_dict[ilu_flag])
                dark_subtracted = DARK_SUB[1]
            else:
                dc = None
                dark_subtracted = DARK_SUB[0]
                logger.warning("dark measurement not subtracted.")
        else:
            dc = None
            dark_subtracted = ""
            logger.warning("lidar " "xarray.Dataset" " not provided.")
    else:
        dc, dark_subtracted = None, ""
    return dc, dark_subtracted
